from .account import ParsingAccount
from vk_parsing.exceptions import *
from vk_parsing.api import VkApiSimple
import os
import datetime
import time
import random

def forward(response):
    return response

class VkApiAccount(ParsingAccount):
    def __init__(self, login, token, cooldown, logger):
        self.vk_session = None
        self.last_access = None
        super().__init__(login, token, cooldown, logger)    

    def auth(self):
        self.vk_session = VkApiSimple(self.token)

    def check(self):
        if (self.vk_session == None):
            raise RuntimeError("Auth() was not called") 
        try:
            self.method("account.getInfo")
            self.can_single = True
        except Exception as ex:
            self.logger.debug(f"{self.login} can_single failed: {str(ex)}")
            self.can_single = False 
        try:
            self.method("execute", {"code": "return [API.account.getInfo()];"})
            self.can_bucket = True
        except Exception as ex:
            self.logger.debug(f"{self.login} can_bucket failed: {str(ex)}")
            self.can_bucket = False
    
    def error_filter(self, ex):
        try:
            if ex.code in [5, 37]: # Errors that should stop parsing
                raise StopParsingError(f"Parsing stopped on account {self.login}, error: {str(ex)}")
            if not ex.code in [6, 9, 14, 29]: # Errors that should drop single account
                raise ex 
        except Exception: # HTTP error, just raise
            raise ex
        return

    def _obey_cooldown(self):
        if (self.last_access == None):
            self.last_access = datetime.datetime.now()
        now_time = datetime.datetime.now()
        difftime = now_time - self.last_access
        if (difftime.total_seconds() < self.cooldown):
            time.sleep(self.cooldown + self._exactly_small_random(self.cooldown) - difftime.total_seconds())
        self.last_access = datetime.datetime.now()

    def _exactly_small_random(self, core_value):
        return core_value / random.randint(1, 100)

    def method(self, method_name, method_args={}, callback=forward, callback_args=(), callback_kwargs={}):
        self._obey_cooldown()
        if (self.vk_session == None):
            raise RuntimeError("Auth() was not called")
        drop = False
        try:
            try:
                vk_response = self.vk_session.method(method_name, params=method_args)
            except Exception as ex:
                self.logger.debug(f"Error with {self.login}: {str(ex)}, trying again")
                self.error_filter(ex)
                self.logger.debug("This error passed filter, dropping account")
                drop = True
        except Exception as ex: # Case we raise the error
            self.logger.debug("This error didn't pass filter, raising this")
            raise
        if drop: # Case we drop the account
            raise BrokenAccountError("")
        try: # Case we call the callback
            return callback(vk_response, *callback_args, **callback_kwargs)
        except Exception as ex:
            self.logger.error(f"Error while calling callback: {str(ex)}")
            raise



class StopParsingError(Exception):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __str__(self):
            return super().__str__()

class BrokenAccountError(Exception):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __str__(self):
            return super().__str__()

class VkApiError(Exception):
    pass

class ApiError(VkApiError):

    def __init__(self, vk, method, values, error):
        super(ApiError, self).__init__()

        self.vk = vk
        self.method = method
        self.values = values
        self.code = error['error_code']
        self.error = error

    def __str__(self):
        return '[{}] {}'.format(self.code, self.error['error_msg'])

class ApiHttpError(VkApiError):

    def __init__(self, vk, method, values, response):
        super(ApiHttpError, self).__init__()
        self.vk = vk
        self.method = method
        self.values = values
        self.response = response

        return self.vk.method(self.method, self.values, raw=self.raw)

    def __str__(self):
        return 'Http error, response code {}'.format(self.response.status_code)



class ParsingTask:
    def __init__(self, method, method_args, callback, callback_args, callback_kwargs):
        self.method = method
        self.method_args = method_args
        self.callback = callback
        self.callback_args = callback_args
        self.callback_kwargs = callback_kwargs



from .accounts.sync_account import VkApiAccount
from .task import ParsingTask
from .lconfig import *
from .exceptions import *
import logging.config

DEFAULT_SLEEP_TIME = 0.4
DEFAULT_BUCKET_SIZE = 25

def forward(response):
    return response

class Parser:
    """
    :param logins: Список логинов ВК (или других произвольных описаний)
    :type logins: list
    
    :param tokens: Список токенов ВК (той же длины)
    :type tokens: list

    :param account_class: Некоторая релизация ParsingAccount
    :type account_class: class
    
    :param sleep_time: Время ожидания между запросами
    :type sleep_time: float

    :param bucket_size_limit: Размер пакета запросов для групповых запросов
    :type bucket_size_limit: float

    :param logger: Логгер, определённый пользователем
    :type logger: logging.Logger

    :param need_logging: Нужно ли логгировать
    :type need_logging: bool

    Два принципа:
    1) Выбрасывание StopParsingError оставляет парсер в невалидном состоянии, ожидается, что работа не будет продолжена
    2) Отложенная задача может быть не выполнена. В том числе, если коллбэк бросает исключение, накрылся интернет или получено StopParsingError 
    """
    
    def __init__(self, logins, tokens, account_class = VkApiAccount, sleep_time = DEFAULT_SLEEP_TIME, bucket_size_limit = DEFAULT_BUCKET_SIZE, logger = None, need_logging = True):
        if not need_logging:
            logging.config.dictConfig(SILENT_CONFIG)
            self.logger = logging.getLogger(list(SILENT_CONFIG["loggers"].keys())[0])
        elif logger == None:
            logging.config.dictConfig(LOGGING_CONFIG)
            self.logger = logging.getLogger(list(LOGGING_CONFIG["loggers"].keys())[0]) 
        else:
            self.logger = logger
        self.logger.debug("Parser was created, logging enabled")
        self.bucket_size_limit = bucket_size_limit
        self.current_account_index = -1
        self.tasks = []
        self.accounts = []
        if bucket_size_limit > DEFAULT_BUCKET_SIZE:
            raise RuntimeError("Too large bucket_size_limit")
        if len(logins) != len(tokens):
            self.logger.error("Weird list lengths, aborting")
            raise RuntimeError("No match login-token: different lengths")
        for login, token in zip(logins, tokens):
            if (not isinstance(login, str)) or (not isinstance(token, str)):
                self.logger.error("Strings were expected, aborting")
                raise RuntimeError("Strings were expected")
        temp_accounts = []
        for login, token in zip(logins, tokens):
            temp_accounts.append(account_class(login, token, sleep_time, self.logger))
        for account in temp_accounts:
            try:
                account.auth()
                account.check()
                self.accounts.append(account)
            except Exception as ex:
                self.logger.warning(f"Account {account.login} auth[or the first request] missed with error: {str(ex)}")
    
    def _get_next_account(self, need_single = False, need_bucket = False):
        """ Выбор аккаунта с необходимыми параметрами

        Бросает `StopParsingError`, если таких аккаунтов нет
        """

        if len(self.accounts) == 0:
            self.logger.error("No more accounts in Parser, aborting")
            raise StopParsingError("No more accounts at all")
        for i in range(len(self.accounts)):
            self.current_account_index += 1
            self.current_account_index %= len(self.accounts)
            account = self._get_current_account()
            acceptable = True
            if (need_single and not account.can_single):
                acceptable = False
            if (need_bucket and not account.can_bucket):
                acceptable = False
            if (acceptable):
                return account
        self.logger.error(f'No more accounts with need_single={need_single}, need_bucket={need_bucket}')
        raise StopParsingError(f'No more accounts with need_single={need_single}, need_bucket={need_bucket}')

    def _get_current_account(self):
        """ Возвращает нынешний аккаунт
        
        Бросает `StopParsingError`, если аккаунтов нет
        """

        if len(self.accounts) == 0:
            self.logger.error("No more accounts in Parser, aborting")
            raise StopParsingError("No more accounts at all")
        return self.accounts[self.current_account_index]

    def _find_account_by_hash(self, hsh):
        """ Возвращает аккаунт по хэшу

        Бросает `RuntimeError`, если такого аккаунта нет
        """

        for account in self.accounts:
            if (hsh == hash(account)):
                return account
        raise RuntimeError("No such account.")

    def _drop_current_account(self):
        """ Удаляет нынешний аккаунт
        
        Бросает `StopParsingError`, если больше аккаунтов нет
        """

        self.logger.debug(f"Erasing {self._get_current_account().login}")
        self.accounts.pop(self.current_account_index)
        if len(self.accounts) == 0:
            self.logger.error("No more accounts in Parser, aborting")
            raise StopParsingError("No more accounts at all")
        self.current_account_index %= len(self.accounts)

    def direct_call(self, method, method_args = {}, callback = forward, callback_args=(), callback_kwargs={}):
        """ Прямой запрос к API с доступного аккаунта
        
        :param method: Название метода API
        :type method: str

        :param method_args: Параметры метода API
        :type method_args: dict
        
        :param callback: Вызываемый коллбэк по выполнении запроса
        :type callback: function

        :param callback_args: Аргументы коллбэка
        :type callback_args: tuple

        :param callback_kwargs: Именованые аргументы коллбэка
        :type callback_kwargs: dict

        Коллбэк вызывается следующим образом: callback(response, *callback_args, **callback_kwargs)
        В случае если callback не передан, то данный метод пробрасывает response напрямую

        Бросает `StopParsingError`, если необходимо
        Бросает ошибки окружения, типа `ApiHttpError` и нестрашные ошибки API: `ApiError`
        Пробрасывает исключения из коллбэка

        В случае успеха возвращает результат коллбэка
        """

        while(True):
            current_account = self._get_next_account(need_single=True, need_bucket=(True if (method == "execute") else False))
            try:
                return current_account.method(method, method_args, callback, callback_args, callback_kwargs)
            except BrokenAccountError as ex:
                self._drop_current_account() # Could raise StopParsingError
                continue

    def add_task(self, method, method_args, callback, callback_args=(), callback_kwargs={}):
        """ Отложенный запрос к API с доступного аккаунта
        
        Сигнатура аналогична сигнатуре метода `direct_call`.
    
        Исполняет execute_tasks, если достигнут `bucket_size_limit`
        Бросает ошибки аналогично методу execute_tasks
        """

        self.tasks += [ParsingTask(method, method_args, callback, callback_args, callback_kwargs)]
        if (len(self.tasks) > self.bucket_size_limit):
            raise StopParsingError("Bucket was somehow overloaded")
        if (len(self.tasks) == self.bucket_size_limit):
            self.execute_tasks()         

    def execute_tasks(self):
        """ Выполняет все существующие задачи
 
        Бросает `StopParsingError`, если необходимо
        Бросает ошибки окружения, типа `ApiHttpError` и нестрашные ошибки API: `ApiError`
        Пробрасывает исключения из коллбэка
        """
        while (len(self.tasks) > 0):
            self._execute_tasks()

    def _execute_tasks(self):
        """ Выполняет существующие задачи
       
        Может не исполнить все задачи, но в таком случае количество аккаунтов будет уменьшено 
        
        Бросает `StopParsingError`, если необходимо
        Бросает ошибки окружения, типа `ApiHttpError` и нестрашные ошибки API: `ApiError`
        Пробрасывает исключения из коллбэков
        """

        if (len(self.tasks) == 0):
            return
        code = "return ["
        for task in self.tasks:
            code += "API." + task.method + "(" + str(task.method_args) + "), " 
        code = code[:-2]
        code += "];"
        try:
            correct = self.direct_call("execute",
                        {"code": code},
                        self._execute_callbacks,
            )
        except Exception as ex:
            self.logger.error(f"Batch direct_call failed, cause: {ex}")
            self.tasks = []
            raise

        if not correct:
            self.logger.debug(f"Trying to understand if account needs mark can_backet=False with direct_calls")
            need_make_impotent = False
            hsh = hash(self._get_current_account())
            while self.tasks:
                task = self.tasks[0]
                self.tasks.pop(0)
                try:
                    self.direct_call(task.method, task.method_args, task.callback, task.callback_args, task.callback_kwargs)
                    need_make_impotent = True
                    break
                except ApiError as ex: # Задача никогда не сможет быть исполнена -- основной случай
                    pass
                except StopParsingError as ex: # Что-то пошло не так
                    self.logger.error(f"Callback threw exception, raising, cause: {ex}")
                    raise
                except ApiHttpError as ex: # Что-то не то с окружением
                    self.logger.error(f"Check direct_call failed, raising, cause: {ex}")
                    self.tasks = []
                    raise 
                except Exception as ex: # Коллбэк бросил исключение
                    self.logger.error(f"Callback threw exception, raising, cause: {ex}")
                    self.tasks = []
                    raise
            if need_make_impotent:
                self.logger.debug("Result: can_bucket=False")
                try:
                    self._find_account_by_hash(hsh).can_bucket = False
                except Exception as ex:
                    self.logger.debug(f"Strange error in execute_tasks: {ex}")
            else:
                self.logger.debug("Result: can_bucket=True")
        else:
            self.tasks = []

    def _execute_callbacks(self, response):
        """ Коллбэк, исполняющий коллбэки tasks
        
        В общем случае имеет право не исполнить некоторый коллбэк.
        Пробрасывает исключение и не выполняет следующие коллбэки, если один коллбэк бросает исключение.
        
        В случае отсутствия исключений
        Возвращает True, если получил хотя бы один ответ (бывают и пропуски, которые на деле False) в групповом ответе.
        Возвращает False иначе. На основе этого делается проверка по запросам в execute_tasks с помощью direct_call.
        """

        size = len(self.tasks)
        for i in range(len(self.tasks)):
            if (isinstance(response[i], bool) and not response[i]):
                size -= 1
        if (size == 0):
            self.logger.debug("Whole bucket request failed (response[i] = False)") 
            return False

        for i in range(len(self.tasks)):
            if (isinstance(response[i], bool) and not response[i]):
                continue
            self.tasks[i].callback(response[i], *self.tasks[i].callback_args, **self.tasks[i].callback_kwargs)
        return True



LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,

    'formatters': {
        'default_formatter': {
            'format': '[%(levelname)s:%(asctime)s] %(message)s'
        },
    },

    'handlers': {
        'default_handler': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'default_formatter'
        }
    },

    'loggers': {
        'vk_parsing': {
            'handlers': ['default_handler'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}

SILENT_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,

    'handlers': {
        'default_handler': {
            'class': 'logging.NullHandler',
            'level': 'DEBUG',
        }
    },

    'loggers': {
        'vk_parsing': {
            'handlers': ['default_handler'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}



class ParsingAccount:
    """
    Абстрактный класс

    :param login: Логин аккаунта ВКонтакте (или любое другое описание аккаунта)
    :type login: str
    
    :param token: Токен аккаунта ВКонтакте
    :type token: str
    
    :param cooldown: Задержка между запросами к api
    :type cooldown: float
    
    :param logger: Логгер, определённый пользователем
    :type logger: logging.Logger
    """

    def __init__(self, login, token, cooldown, logger):
        self.login = login
        self.token = token
        self.cooldown = cooldown
        self.logger = logger
        self.can_single = False
        self.can_bucket = False

    def __hash__(self):
        return hash((self.login, self.token))

    def auth(self):
        """ Аутентификация аккаунта.

        Метод, вызывающийся для аккаунта один раз.
        Бросает исключение, если авторизация невозможна.
        """
        pass

    def check(self):
        """ Проверка аккаунта

        Метод, устанавливающий `can_single` и `can_bucket`
        Где
        `can_single` - может ли аккаунт выполнять единичные запросы 
        `can_bucket` - может ли аккаунт выполнять групповые запросы
        """
        pass

    def error_filter(self, error):
        """ Фильтр ошибок, бросаемых обращением к API

        :param error: Фильтруемая ошибка
        :type error: Exception

        Не бросает исключение, если после получения данной ошибки аккаунт нельзя использовать
        Бросает исключение, если ошибка не зависит от аккаунта, в том числе
            `StopParsingError`, если ошибка особо опасна
            Любое другое исключение, если ошибка относится к окружению (например `ApiHttpError`)
        """
        pass
 
    def method(self, method_name, method_args, callback, callback_args, callback_kwargs):
        """ Вызов метода API
        
        :param method: Название метода API
        :type method: str

        :param method_args: Параметры метода API
        :type method_args: dict
        
        :param callback: Вызываемый коллбэк по выполнении запроса
        :type callback: function

        :param callback_args: Аргументы коллбэка
        :type callback_args: tuple

        :param callback_kwargs: Именованые аргументы коллбэка
        :type callback_kwargs: dict

        Коллбэк вызывается следующим образом: callback(response, *callback_args, **callback_kwargs)
 
        Используя `error_filter`:
        Бросает `StopParsingError` в случае необходимости остановить парсинг.
        Бросает `BrokenAccountError` в случае необходимости остановить парсинг на этом аккаунте.
        Пробрасывает исключение в случае ошибки при вызове callback-а
        Пробрасывает исключение в случае ошибки окружения (например `ApiHttpError` и нестрашные ошибки API: `ApiError`)        

        Возвращает `response['response']` в случае успеха
        """
        pass



import requests
import json
import threading
from .exceptions import *

class VkApiSimple:
    """
    :param token: Токен аккаунта ВК
    :type token: str
    
    :param v: Версия используемого api
    :type v: int
    """

    def __init__(self, token, v=5.131):
        self.token = token
        self.v = v
        self.lock = threading.Lock()

    def method(self, method, params={}):
        """ Вызов метода API

        :param method: Название метода
        :type method: str
        
        :param params: Параметры
        :type params: dict

        Бросает `ApiHttpError` в случае неуспешного кода возварата http запроса
        
        Бросает `ApiError` в случае неуспешного запроса к API

        Возвращает `response['response']` в случае успеха
        """

        params = params.copy() if params else {}
        if 'v' not in params:
            params['v'] = self.v

        params["access_token"] = self.token

        with self.lock:
            response = requests.get(
                f"https://api.vk.com/method/{method}",
                params=params
            )

        if response.ok:
            response = response.json()
        else:
            raise ApiHttpError(self, method, params, response)

        if 'error' in response:
            raise ApiError(self, method, params, response['error'])

        return response['response']
